#/*
# * Copyright (c) 2020 - 2025 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */
import sys
import os
import re
import yaml
import logging
import common_utils as utils

from lopper.log import _init, _warning, _info, _error, _debug, _level, __logger__

sys.path.append(os.path.dirname(__file__))

_init(__name__)

def get_label(sdt, symbol_node, node):
    prop_dict = symbol_node.__props__
    match = [label for label,node_abs in prop_dict.items() if re.match(node_abs[0], node.abs_path) and len(node_abs[0]) == len(node.abs_path)]
    if match:
        return match[0]
    elif node.propval('xlnx,name') != ['']:
        return node.propval('xlnx,name', list)[0]
    else:
        return None

def get_cpu_node(sdt, options):
    cpu_name = options['args'][0]
    symbol_node = sdt.tree['/__symbols__']
    nodes = sdt.tree.nodes('/cpu.*')
    cpu_labels = []
    matched_label = None
    for node in nodes:
        matched_label = get_label(sdt, symbol_node, node)
        if matched_label is not None:
            if matched_label == cpu_name:
                return node
            elif node.propval('reg') != ['']:
                cpu_labels.append(matched_label)
    _error(f"In valid CPU Name valid Processors for a given SDT are {cpu_labels}\n")
    sys.exit(1)

def item_generator(json_input, lookup_key):
    if isinstance(json_input, dict):
        for k, v in json_input.items():
            if k == lookup_key:
                if isinstance(v, str):
                    yield [v]
                else:
                    yield v
            else:
                for child_val in item_generator(v, lookup_key):
                    yield child_val
    elif isinstance(json_input, list):
        for item in json_input:
            for item_val in item_generator(item, lookup_key):
                yield item_val

# This API reads the schema and returns the compatible list
def compat_list(schema):
    compatible_list = []
    if schema and 'compatible' in schema.get('properties',{}).keys():
        sch = schema['properties']['compatible']
        compatible_list = []
        for l in item_generator(sch, 'enum'):
            compatible_list.extend(l)

        for l in item_generator(sch, 'const'):
            compatible_list.extend(l)

        if 'contains' in sch.keys():
            for l in item_generator(sch['contains'], 'enum'):
                compatible_list.extend(l)

            for l in item_generator(sch['contains'], 'const'):
                compatible_list.extend(l)
        compatible_list = list(set(compatible_list))
    return compatible_list

"""
This API scans the device-tree node and returns the address
and size of the reg property for the user provided index.

Args:
    node: LopperNode object
    value: property value
    idx: index
"""
def scan_reg_size(node, value, idx):
    na = node.parent["#address-cells"].value[0]
    ns = node.parent["#size-cells"].value[0]
    cells = na + ns
    reg = 0
    size = 0
    if cells > 2:
        try:
            reg1 = value[cells * idx]
            reg = int(f"{hex(reg1)}{value[cells * idx + 1]:08x}", base=16) if reg1 !=0 else value[cells * idx + 1]
            size1 = value[cells * idx + na]
            size = int(f"{hex(size1)}{value[cells * idx + ns + 1]:08x}", base=16) if size1 !=0 else value[cells * idx + ns + 1]
        except IndexError:
            _debug(f"The Reg Property is corrupted for the {node.name} and value is {value}")
            pass
    elif cells == 2:
        try:
            reg = value[idx * cells]
            size = value[idx * cells + 1]
        except IndexError:
            _debug(f"The Reg Property is corrupted for the {node.name} and value is {value}")
            pass
    else:
        try:
            reg = value[0]
        except IndexError:
            _debug(f"The Reg Property is corrupted for the {node.name} and value is {value}")
            pass

    # 4 GB (0xFFFFFFFF) is a boundary condition for a 32-bit cell size. Size is calculated
    # using (high_addr - low_addr + 1) in SDT. When the complete 4 GB DDR is used for a
    # 32 bit processor, the size becomes 0x100000000 using the above calculation and result
    # into an overflow for a 32-bit cell size in SDT. To avoid this overflow, (+1) is
    # removed in SDT for this particular boundary condition and it is being done here in
    # assist to get the correct size in linker and xparameters.
    if size == 4294967295 and reg == 0:
        size = 4294967296

    return reg, size

def get_interrupt_prop(sdt, node, value):
    intr = []
    inp =  node['interrupt-parent'].value[0]
    intr_parent = [node for node in sdt.tree['/'].subnodes() if node.phandle == inp]
    inc = intr_parent[0]["#interrupt-cells"].value[0]
    """
    Baremetal Interrupt Property format:
        bits[11:0]  interrupt-id
        bits[15:12] trigger type and level flags
        bits[19:16] CPU Mask
        bit[20] interrupt-type (1: PPI, 0: SPI)
    """
    # Below logic converts the interrupt propery value to baremetal
    # interrupt property format.
    nintr = len(value)/inc
    tmp = inc % 2
    for val in range(0, int(nintr)):
        intr_sensitivity = value[tmp+1] << 12
        intr_id = value[tmp]
        # Convert PPI interrupt to baremetal interrupt format
        if value[tmp-1] == 1 and inc == 3:
            intr_sensitivity = (value[tmp+1] & 0xF) << 12
            cpu_mask = (value[tmp+1] & 0xFF00) << 8
            ppi_type = 1 << 20
            intr_id = intr_id + cpu_mask + ppi_type
        intr.append(hex(intr_id + intr_sensitivity))
        tmp += inc

    return intr

def get_interrupt_id(sdt, node, value):
    intr = []
    inp =  node['interrupt-parent'].value[0]
    intr_parent = [node for node in sdt.tree['/'].subnodes() if node.phandle == inp]
    inc = intr_parent[0]["#interrupt-cells"].value[0]
    nintr = len(value)/inc
    tmp = inc % 2
    for val in range(0, int(nintr)):
        intr_id = value[tmp]
        intr.append(int(hex(intr_id), 16))
        tmp += inc
    return intr


#Return the base address of the parent node.
def get_phandle_regprop(sdt, prop, value):
    parent_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    reg, size = scan_reg_size(parent_node[0], parent_node[0]['reg'].value, 0)
    # Special handling for Soft Ethernet(1/2.5G, and 10G/25G MAC) axistream-connected property
    if prop == "axistream-connected":
        compat = parent_node[0].propval('compatible')
        axi_fifo = [item for item in compat if "xlnx,axi-fifo" in item]
        axi_dma = [item for item in compat if "xlnx,eth-dma" in item]
        axi_mcdma = [item for item in compat if "xlnx,eth-mcdma" in item]
        if axi_fifo:
            reg += 1
        elif axi_dma:
            reg += 2
        elif axi_mcdma:
            reg += 3
    return reg

#Return the base address of the interrupt parent.
def get_intrerrupt_parent(sdt, value):
    intr_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    reg, size = scan_reg_size(intr_node[0], intr_node[0]['reg'].value, 0)
    """
    Baremetal Interrupt Parent Property Format:
        bits[0]    Interrupt parent type (0: GIC, 1: AXI INTC)
        bits[31:1] Base Address of the interrupt parent
    """
    compat = intr_node[0].propval('compatible')
    axi_intc = [item for item in compat if "xlnx,xps-intc-1.00.a" in item]
    if axi_intc:
        reg += 1
    return reg

"""
This API scans the pci node's range property and returns the required address
and size.

Args:
    range_value: List containing one range values.
    ns: size cells value for the pci node.
"""
def scan_ranges_size(range_value, ns):
    """
    Both address and size can be either 32 bit or 64 bit.
    Check the higher cell for both and if that is not zero,
    concatenate the higher and the lower cells.
    e.g. <0x4 0x80000000> => 0x480000000
    """
    addr = range_value[2]
    high_addr_cell = range_value[1]
    if high_addr_cell != 0:
        addr = int(f"{hex(high_addr_cell)}{addr:08x}", base=16)

    size = range_value[-1]
    high_size_cell = range_value[-2]
    # If ns = 1, then there is no use of high_size_cells
    if high_size_cell != 0 and ns > 1:
        size = int(f"{hex(high_size_cell)}{size:08x}", base=16)

    return addr, size

def get_clock_offset(node):
    if node.propval("clock-names") != ['']:
        clock_names = node["clock-names"].value
        try:
            tclk_offset = clock_names.index("tx_clk")
            namelen = len(node["clock-names"].value)
            clocklen = len(node["clocks"].value)
            if namelen != clocklen:
                tclk_offset = 2 * tclk_offset + 1
                return tclk_offset
            return tclk_offset
        except Exception:
            return 1

    try:
        ncells = len(node["clocks"].value)
    except KeyError:
        return 1
    if ncells <= 1:
        return 0
    return 1

def get_clock_prop(sdt, value, offset):
    """
    Baremetal clock format:
        bits[0] clock parent(controller) type(0: ZynqMP clock controller)
        bits[31:1] clock value
    """
    if offset < len(value):
        return value[offset]
    if not value:
        return 0
    if len(value) > 1:
        _warning(f"Invalid clock offset returning first clock")
    return value[0]


def get_pci_ranges(node, value, pad):
    """
    For pci range, address cells is always 3, size-cells may vary between 1 or 2 (usually it is 2).
    Sample pci range for address-cells = 3 and size-cells = 2:
    eg.1: <0x43000000 0x00000080 0x00000000 0x00000080 0x00000000 0x00000000 0x80000000>
    eg.2: <0x42000000 0 0x80000000  0x80000000  0 0x20000000>
    cell[0] => chip-select (Flag for 32bit or 64 bit PCIe Bar)
    cell[1] cell[2] => PCIe address (needed in the config structure)
    cell[-2] cell[-1] => Size of the PCIe Bus (needed to get High address in config table)
    Middle cells (cell[3] cell[4] => In example 1, cell[3] => in example 2)
      => The CPU host address where the PCIe address will be mapped
      => The middle cell size (i.e. 2 or 1) depends on the SoC bus width.
    """
    na = node["#address-cells"].value[0]
    ns = node["#size-cells"].value[0]
    """
    Probable combinations for pci range are:
    <3 address cells> <2 cpu host addr cell> <2 size cells>
    <3 address cells> <1 cpu host addr cell> <2 size cells>
    <3 address cells> <1 cpu host addr cell> <1 size cells>

    cell_size = na + (2*ns) covers first and third conditions
    if condition covers the second one.
    """
    cell_size = na + (2 * ns)

    if len(value) % 6 == 0:
        cell_size = 6

    pci_range_split = []
    for i in range(pad):
        pci_range_split.append(value[i*cell_size:(i+1)*cell_size])

    pci_ranges = []
    for ranges in pci_range_split:
        if ranges:
            reg, size = scan_ranges_size(ranges, ns)
            high_addr = reg + size - 1
            pci_ranges += [hex(reg), hex(high_addr)]
        else:
            pci_ranges += [hex(0), hex(0)]
    return pci_ranges

class DtbtoCStruct(object):
    def __init__(self, out_file):
        self._outfile_path = out_file
        self._lines = []
        # Clear the file initially
        open(self._outfile_path, "w").close()

    def out(self, line):
        """Output a string to the output file using append mode

        Args:
            line: String to output
        """
        with open(self._outfile_path, "a") as f:
            f.write(line)

    def buf(self, line):
        """Buffer up a string to send later

        Args:
            line: String to add to our 'buffer' list
        """
        self._lines.append(line)

    def get_buf(self):
        """Get the contents of the output buffer, and clear it

        Returns:
            The output buffer, which is then cleared for future use
        """
        lines = self._lines
        self._lines = []
        return lines


def is_compat(node, compat_string_to_test):
    if re.search( "module,baremetalconfig_xlnx", compat_string_to_test):
        return xlnx_generate_bm_config
    return ""

def get_stdin(sdt, chosen_node, node_list):
    if chosen_node.propval('stdout-path') != ['']:
        prop_val = chosen_node['stdout-path'].value
        serial_node = sdt.tree.alias_node(prop_val[0].split(':')[0])
        if serial_node:
            match = [x for x in node_list if re.search(x.name, serial_node.name)]
            return match[0]
    return 0

def get_mapped_nodes(sdt, node_list, options):
    # Yocto Machine to CPU compat mapping
    match_cpunode = get_cpu_node(sdt, options)

    all_phandles = []
    address_map = match_cpunode.parent["address-map"].value
    na = match_cpunode.parent["#ranges-address-cells"].value[0]
    ns = match_cpunode.parent["#ranges-size-cells"].value[0]
    cells = na + ns
    tmp = na
    while tmp < len(address_map):
        all_phandles.append(address_map[tmp])
        tmp = tmp + cells + na + 1

    valid_nodes = [node for node in node_list for handle in all_phandles if handle == node.phandle]
    return valid_nodes

def xlnx_generate_config_struct(sdt, node, drvprop_list, plat, driver_proplist, is_subnode, options):
    for i, prop in enumerate(driver_proplist):
        pad = 0
        phandle_prop = 0
        subnode_gen = 0
        # Few drivers has multiple data interface type (AXI4 or AXI4-lite),
        # Driver config structures of these SoftIP's contains baseaddress entry for each possible data interface type.
        # Device-tree node reg property may or may not contain all the possible entries that driver config structure
        # is expecting, In that case we need to add dummy entries(0xFF) in the config structure in order to avoid
        # compilation errors.
        #
        # Yaml meta-data representation/syntax will be like below
        # reg: <range of baseaddress>
        # interrupts: <supported range of interrupts>
        if isinstance(prop, dict):
            pad = list(prop.values())[0]
            prop = list(prop.keys())[0]
            if isinstance(pad, dict):
                subnode_prop = prop
                prop = list(pad.keys())[0]
                subnode_prop_list = list(pad.values())[0]
                if prop == "subnode_phandle":
                    subnode_gen = 1
                    pad = 0
                    phandle_prop = 0
            if pad == "phandle":
                phandle_prop = 1
    
        if i == 0 and is_subnode:
            plat.buf('\n\t\t{')
        elif i == 0:
            plat.buf('\n\t{')

        if not subnode_gen:
            xlnx_generate_prop(sdt, node, prop, drvprop_list, plat, pad, phandle_prop, options)
        else:
            # Get the sub node
            dummy_struct = None
            try:
                phandle_value = node[subnode_prop].value[0]
            except KeyError:
                # Need to create dummy entries
                _debug(f"Find sub node is failed for {node.name}, creating dummy structure")
                dummy_struct = True

            if dummy_struct:
                for k,subnode_prop in enumerate(subnode_prop_list):
                    if k == 0:
                        plat.buf('\n\t\t{')
                    plat.buf('\n\t\t%s' % hex(0x00))
                    if k == len(subnode_prop_list)-1:
                        plat.buf('\n\t\t}')
                    else:
                        plat.buf(',')
            else:
                sub_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == phandle_value]
                xlnx_generate_config_struct(sdt, sub_node[0], drvprop_list, plat, subnode_prop_list, 1, options)
    
        if i == len(driver_proplist)-1:
            if "subnode_phandle" not in prop:
                plat.buf(' /* %s */' % prop)
            if is_subnode:
                plat.buf('\n\t\t}')
            else:
                plat.buf('\n\t},')
        else:
            plat.buf(',')
            if "subnode_phandle" not in prop:
                plat.buf(' /* %s */' % prop)

def xlnx_generate_prop(sdt, node, prop, drvprop_list, plat, pad, phandle_prop, options):
    nosub = 0
    if prop == "reg":
        val, size = scan_reg_size(node, node[prop].value, 0)
        drvprop_list.append(hex(val))
        plat.buf('\n\t\t%s' % hex(val))
        if pad:
            for j in range(1, pad):
                try:
                    val, size = scan_reg_size(node, node[prop].value, j)
                    drvprop_list.append(hex(val))
                    plat.buf(',\n\t\t%s' % hex(val))
                except IndexError:
                    _debug(f"Find reg size is failed for {node.name}, adding default value as {hex(0xFFFF)}")
                    plat.buf(',\n\t\t%s' % hex(0xFFFF))
    elif prop == "compatible":
        plat.buf('\n\t\t%s' % '"{}"'.format(node[prop].value[0]))
        drvprop_list.append(node[prop].value[0])
    elif prop == "interrupts":
        try:
            intr = get_interrupt_prop(sdt, node, node[prop].value)
        except KeyError:
            _debug(f"Get interrupt prop is failed {node.name}, adding default value as {hex(0xFFFF)}")
            intr = [hex(0xFFFF)]

        if pad:
            plat.buf('\n\t\t{')
            for j in range(0, pad):
                try:
                    plat.buf('%s' % intr[j])
                    drvprop_list.append(intr[j])
                except IndexError:
                    _debug(f"Get interrupt prop is failed {node.name}, adding default value as {hex(0xFFFF)}")
                    plat.buf('%s' % hex(0xFFFF))
                    drvprop_list.append(hex(0xFFFF))
                if j != pad-1:
                    plat.buf(',  ')
            plat.buf('}')
        else:
            plat.buf('\n\t\t%s' % intr[0])
            drvprop_list.append(intr[0])
    elif prop == "interrupt-parent":
        try:
            intr_parent = get_intrerrupt_parent(sdt, node[prop].value)
        except KeyError:
            _debug(f"Get interrupt parent is failed {node.name}, adding default value as {hex(0xFFFF)}")
            intr_parent = 0xFFFF
        plat.buf('\n\t\t%s' % hex(intr_parent))
        drvprop_list.append(hex(intr_parent))
    elif prop == "clocks":
        try:
            tclk_offset = get_clock_offset(node)
            clkprop_val = get_clock_prop(sdt, node[prop].value, tclk_offset)
        except KeyError:
            _warning(f"Clock property is missing for this node {node.name}, adding default value as {hex(0xFFFF)}")
            clkprop_val = 0xFFFF
        plat.buf('\n\t\t%s' % hex(clkprop_val))
        drvprop_list.append(hex(clkprop_val))
    elif prop == "mdioproducer-baseaddr":
        if node.propval('phy-handle') == ['']:
            drvprop_list.append(hex(0))
            plat.buf('\n\t\t%s' % hex(0))
        else:
            # two possible scenarios phy-handle property is part of mac node or part of mdio node handle both the cases
            parent_node = [node1.parent for node1 in sdt.tree['/'].subnodes() if node1.phandle == node['phy-handle'].value[0]]
            if parent_node:
                # if parent node has reg property it's mac node other wise its mdio node
                if parent_node[0].propval('reg') != ['']:
                    reg, size = scan_reg_size(parent_node[0], parent_node[0]['reg'].value, 0)
                    drvprop_list.append(hex(reg))
                    plat.buf('\n\t\t%s' % hex(reg))
                elif parent_node[0].parent.propval('reg') != ['']:
                    reg, size = scan_reg_size(parent_node[0].parent, parent_node[0].parent['reg'].value, 0)
                    drvprop_list.append(hex(reg))
                    plat.buf('\n\t\t%s' % hex(reg))
            else:
                drvprop_list.append(hex(0))
                plat.buf('\n\t\t%s' % hex(0))
    elif prop == "gmiitorgmii-addr":
      gmii2rgmii_addr = 0  # Default value
      # Check if any subnode has the compatible property 'xlnx,gmii-to-rgmii-1.0'
      for subnode in node.subnodes():
          if  subnode.propval('compatible') and "xlnx,gmii-to-rgmii-1.0" in subnode.propval('compatible'):
              reg, size = scan_reg_size(subnode, subnode['reg'].value, 0)
              gmii2rgmii_addr = reg

      drvprop_list.append(hex(gmii2rgmii_addr))
      plat.buf('\n\t\t%s' % hex(gmii2rgmii_addr))
    elif prop == "child,required":
        plat.buf('\n\t\t{')
        numsplits = 0
        splitidx = 0
        for j,child in enumerate(list(node.child_nodes.items())):
            # Avoid sysmon temp child node as it is not supported in baremetal
            if re.search(r"(sysmon@[a-zA-Z0-9]+/temp@[a-zA-Z0-9]+$)", child[0]):
                continue
            if len(pad) != 1:
                plat.buf('\n\t\t\t{')
            for k,p in enumerate(pad):
                if p == "nosub":
                    nosub = 1
                if len(p) == 1:
                    if 'split' in p.keys():
                        splitidxs = []
                        for idx in str(p['split']).split(' '):
                            splitidxs.append(int(idx))
                        numsplits = len(splitidxs);
                        continue
                    pv = list(p.values())
                    xlnx_generate_prop(sdt, child[1], prop, drvprop_list, plat, pv[0], phandle_prop, options)
                else:
                    try:
                        plat.buf('\n\t\t\t\t%s' % str(child[1][p].value[0]))
                        drvprop_list.append(str(child[1][p].value[0]))
                        if k != (len(pad) - 1) or len(pad) == 1:
                            plat.buf(',')
                        plat.buf(' /* %s */' % p)
                    except KeyError:
                        if nosub == 0:
                            _debug(f"Failed to find the driver prop list for {node.name}, adding default value as {hex(0xFFFF)}")
                            plat.buf('\n\t\t\t\t%s' % hex(0xFFFF))
                            drvprop_list.append(hex(0xFFFF))
            if len(pad) != 1:
                if j != (len(list(node.child_nodes.items())) - 1):
                    plat.buf('\n\t\t\t},')
                else:
                    plat.buf('\n\t\t\t},')
            if numsplits:
                if splitidx < numsplits:
                    if j == splitidxs[splitidx]:
                        plat.buf('\n},\n{')
                        splitidx = splitidx + 1
        plat.buf('\n\t\t},')
    elif phandle_prop:
        try:
            prop_val = get_phandle_regprop(sdt, prop, node[prop].value)
        except KeyError:
            _debug(f"Get phandle reg property failed for {node.name}, adding default value as {0}")
            prop_val = 0
        plat.buf('\n\t\t%s' % hex(prop_val))
        drvprop_list.append(hex(prop_val))
    elif prop == "ranges":
        try:
            device_type = node['device_type'].value[0]
            if device_type == "pci":
                device_ispci = 1
        except KeyError:
            _debug(f"Get device type failed for {node.name}, adding default value as {0}")
            device_ispci = 0
        if device_ispci:
            match_cpunode = get_cpu_node(sdt, options)
            if match_cpunode.propval('xlnx,ip-name') != ['']:
                ip_name = match_cpunode.propval('xlnx,ip-name', list)[0]
                if ip_name in ["psu_cortexr5", "psv_cortexr5", "psx_cortexr52", "cortexr52"]:
                    pad = 1
            prop_vallist = get_pci_ranges(node, node[prop].value, pad)
            for j, prop_val in enumerate(prop_vallist):
                plat.buf('\n\t\t%s' % prop_val)
                if j != (len(prop_vallist) - 1):
                    plat.buf(',')
                drvprop_list.append(prop_val)
        # Add special handling for Zynq NAND
        elif "arm,pl353-smc-r2p1" in node.propval('compatible'):
            plat.buf('\n\t\t%s' % hex(node[prop].value[2]))
            drvprop_list.append(hex(node[prop].value[2]))

    elif prop == "Handler-table":
        # In the driver config structure this property is of type struct
        # to avoid Missing braces around initializer warning add braces
        # around value (https://gcc.gnu.org/bugzilla/show_bug.cgi?id=53119)
        plat.buf('\n\t\t{{0U}}')
    else:
        try:
            prop_val = node[prop].value
            # For boolean property if present LopperProp will return
            # empty string convert it to baremetal config struct expected value
            if '' in prop_val:
                prop_val = [1]
        except KeyError:
            _debug(f"Get property value failed for {node.name}, adding default value as {0}")
            prop_val = [0]

        if pad:
            address_prop = ""
            for index in range(0,pad):
                if index == 0:
                    address_prop = hex(node[prop].value[index])
                elif index < len(node[prop].value):
                    address_prop += f"{node[prop].value[index]:08x}"
            if address_prop:
                plat.buf(f'\n\t\t{address_prop}')
                drvprop_list.append(address_prop)
        else:
            if ('{' in str(prop_val[0])):
                    prop_val_temp = []
                    for k in (str(prop_val[0]).split()[1]).split(","):
                        prop_val_temp.append(int(k, 16))
                    prop_val = prop_val_temp

            if ('/bits/' in prop_val):
                prop_val = [int(prop_val[-1][3:-1], base=16)]

            if isinstance(prop_val[0], str):
                if prop_val[0].replace('.','',1).isdigit():
                    plat.buf('\n\t\t%s' % '{}'.format(node[prop].value[0]))
                else:
                    plat.buf('\n\t\t%s' % '"{}"'.format(node[prop].value[0]))
                drvprop_list.append(node[prop].value[0])
            elif len(prop_val) > 1:
                plat.buf('\n\t\t{')
                for k,item in enumerate(prop_val):
                    if isinstance(item, int):
                        drvprop_list.append(hex(item))
                    else:
                        drvprop_list.append(item)
                    plat.buf('%s' % item)
                    if k != len(prop_val)-1:
                        plat.buf(',  ')
                plat.buf('}')
            else:
                drvprop_list.append(hex(prop_val[0]))
                plat.buf('\n\t\t%s' % hex(prop_val[0]))


# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal driver meta-data file path
def xlnx_generate_bm_config(tgt_node, sdt, options):
    _level(utils.log_setup(options), __name__)
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    node_list = []
    chosen_node = ""
    if options.get('outdir', {}):
        sdt.outdir = options['outdir']
    # Traverse the tree and find the nodes having status=ok property
    for node in root_sub_nodes:
        try:
            if node.name == "chosen":
                chosen_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
        except:
           pass
    src_dir = options['args'][1]
    stdin = ""
    try:
        stdin = options['args'][2]
        stdin_node = get_stdin(sdt, chosen_node, node_list)
    except IndexError:
        pass

    drvpath = utils.get_dir_path(src_dir.rstrip(os.sep))
    drvname = utils.get_base_name(drvpath)
    # Incase of versioned driver strip the version info
    drvname = re.split(r"_v(\d+)_(\d+)", drvname)[0]
    yaml_file = os.path.join(drvpath, f"data/{drvname}.yaml")
    if not utils.is_file(yaml_file):
        _warning(f"{drvname} Driver doesn't have yaml file")
        return False

    driver_compatlist = []
    # Read the yaml file and get the driver supported compatible list
    # and config data file required driver properties
    schema = utils.load_yaml(yaml_file)
    driver_compatlist = compat_list(schema)
    driver_proplist = schema.get('required',[])
    config_struct = schema.get('config',[])
    driver_optproplist = schema.get('optional',[])

    if driver_proplist == []:
        return True
    driver_nodes = []
    for node in node_list:
        for compat in driver_compatlist:
            try:
                compat_string = node['compatible'].value
                for compa in compat_string:
                    if compat in compa:
                        if not node in driver_nodes:
                            driver_nodes.append(node)
            except KeyError:
                _warning(f"Node {node.name} does not have 'compatible' property")
                pass

    if sdt.tree[tgt_node].propval('pruned-sdt') == ['']:
        driver_nodes = get_mapped_nodes(sdt, driver_nodes, options)
    if not config_struct:
        config_struct = str("X") + drvname.capitalize() + str("_Config")
    else:
        config_struct = config_struct[0]

    out_g_file_name = utils.find_files("*_g.c", src_dir)
    if out_g_file_name:
        out_g_file_name = utils.get_base_name(out_g_file_name[0])
        drvname = out_g_file_name.replace('_g.c','')
    else:
        drvname = config_struct.split('_Config')[0].lower()
        drvname = f"x{drvname[1:]}"
        out_g_file_name = f"{drvname}_g.c"
    outfile = os.path.join(sdt.outdir, out_g_file_name)

    if driver_nodes == []:
        return True

    plat = DtbtoCStruct(outfile)
    nodename_list = []
    for node in driver_nodes:
        nodename_list.append(node.name)

    cmake_file = drvname[1:].upper() + str("Config.cmake")
    cmake_file = os.path.join(sdt.outdir,f"{cmake_file}")
    with open(cmake_file, 'a') as fd:
       fd.write("set(DRIVER_INSTANCES %s)\n" % utils.to_cmakelist(nodename_list))
       if stdin:
           if stdin_node:
               match = [x for x in nodename_list if re.search(x, stdin_node.name)]
               if match:
                   fd.write("set(STDIN_INSTANCE %s)\n" % '"{}"'.format(match[0]))

    for index,node in enumerate(driver_nodes):
        drvprop_list = []
        drvoptprop_list = []
        if index == 0:
            plat.buf('#include "%s.h"\n' % drvname)
            plat.buf('\n%s %s __attribute__ ((section (".drvcfg_sec"))) = {\n' % (config_struct, config_struct + str("Table[]")))
        xlnx_generate_config_struct(sdt, node, drvprop_list, plat, driver_proplist, 0, options)
        if index == len(driver_nodes)-1:
            plat.buf('\n\t {\n\t\t NULL\n\t}')
            plat.buf('\n};')

        for i, prop in enumerate(driver_optproplist):
            if isinstance(prop, dict):
               pad = list(prop.values())[0]
               prop = list(prop.keys())[0]
            if prop == "child,required":

                for j,child in enumerate(list(node.child_nodes.items())):
                    for k,p in enumerate(pad):
                        # Avoid sysmon temp child node as it is not supported in baremetal
                        if not re.search(r"(sysmon@[a-zA-Z0-9]+/temp@[a-zA-Z0-9]+$)", child[0]):
                            drvoptprop_list.append(child[1][p].value[0])
            else:
                try:
                    v0 = node[prop].value[0]
                    drvoptprop_list.append(hex(int(v0.partition(",")[0].strip(), 0) if isinstance(v0, str) else v0 if isinstance(v0, int) else int(v0, 0)))
                except (KeyError, IndexError, TypeError, ValueError):
                    pass

        with open(cmake_file, 'a') as fd:
           fd.write("set(DRIVER_PROP_%s_LIST %s)\n" % (index, utils.to_cmakelist(drvprop_list)))
           fd.write("set(DRIVER_OPTPROP_%s_LIST %s)\n" % (index, utils.to_cmakelist(drvoptprop_list)))
           fd.write("list(APPEND TOTAL_DRIVER_PROP_LIST DRIVER_PROP_%s_LIST)\n" % index)
    plat.out(''.join(plat.get_buf()))

    return True



#/*
# * Copyright (C) 2022 - 2025 Advanced Micro Devices, Inc.  All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.kedareswara.rao@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */
import sys
import os
import glob
import yaml
import re
import common_utils as utils

from baremetalconfig_xlnx import get_cpu_node
from lopper.log import _init, _warning, _info, _error, _debug, _level, __logger__

sys.path.append(os.path.dirname(__file__))

_init(__name__)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetal_getsupported_comp_xlnx", compat_string_to_test):
        return xlnx_baremetal_getsupported_comp
    return ""

class VerboseSafeDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True

def get_yaml_data(comp_name, comp_dir,proc_ip_name=None,family=None,variant=None):
    yaml_file = os.path.join(comp_dir, 'data', f'{comp_name}.yaml')
    schema = utils.load_yaml(yaml_file)
    supported_proc_list = schema.get('supported_processors',[])
    condition_item = next((proc for proc in supported_proc_list if isinstance(proc, dict) and "condition" in proc), None)
    if condition_item:
        supported_proc_list = utils.run_exec(condition_item["condition"], proc_ip_name, family, variant, return_list="supported_processors", yaml_file=yaml_file)
    supported_os_list = schema.get('supported_os',[])
    description = schema.get('description',"")
    dep_lib_list = list(schema.get('depends_libs',{}).keys())
    examples = schema.get('examples', {})
    example_dict = {}
    if examples and proc_ip_name:
        if examples.get("condition"):
            local_scope = utils.run_exec(examples["condition"], proc_ip_name, family, variant, return_list="examples", yaml_file=yaml_file)
            examples = {key: value for key, value in examples.items() if key in local_scope}
        examples.pop("condition",None)
        for ex,deps in examples.items():
            if deps:
                # Read the supported_platforms check if any
                dep_plat_list = [dep for dep in deps if "supported_platforms" in dep]
                dep_file_list = [dep for dep in deps if "dependency_files" in dep]
                if dep_plat_list:
                    plat_list = dep_plat_list[0]['supported_platforms']
                    family_variant = []
                    if family:
                        family_variant.append(family)
                    if variant:
                        family_variant.extend(variant)
                    result = any(item in plat_list for item in family_variant)
                    if result:
                        if dep_file_list:
                            example_dict.update({ex:dep_file_list[0]['dependency_files']})
                        else:
                            example_dict.update({ex:[]})
                elif dep_file_list:
                    example_dict.update({ex:dep_file_list[0]['dependency_files']})
            else:
                example_dict.update({ex:[]})
    return supported_proc_list, supported_os_list, description, dep_lib_list, example_dict

def xlnx_baremetal_getsupported_comp(tgt_node, sdt, options):
    _level(utils.log_setup(options), __name__)
    proc_name = options['args'][0]
    repo_path_data = utils.get_abs_path(options['args'][1])

    matched_node = get_cpu_node(sdt, options)
    if not matched_node:
        _error("No matching CPU node found.")

    proc_ip_name = matched_node['xlnx,ip-name'].value[0]
    family = sdt.tree['/'].propval('family')
    family = family[0] if family else ""
    variant = sdt.tree['/'].propval('variant')
    supported_app_dict = {proc_name: {'standalone': {}, 'freertos': {}}}
    supported_libs_dict = {proc_name: {'standalone': {}, 'freertos': {}}}

    if utils.is_file(repo_path_data):
        path_schema = utils.load_yaml(repo_path_data)
    else:
        path_schema = {
            'library' : {},
            'apps'    : {}
        }

        files = glob.glob(repo_path_data + '/**/data/*.yaml', recursive=True)
        if not files:
            _warning(f"No YAML files found in {repo_path_data}.")

        for entries in files:
            dir_path = utils.get_dir_path(utils.get_dir_path(entries))
            comp_name = utils.get_base_name(dir_path)
            comp_name = re.split(r"_v(\d+)_(\d+)", comp_name)[0]
            yaml_data = utils.load_yaml(entries)
            version = yaml_data.get('version','vless')

            if yaml_data['type'] not in ['library','apps']:
                continue

            if yaml_data['type'] in ['library'] and version == 'vless':
                _error(f"""
                    Couldnt set the paths correctly.
                    {comp_name} in {repo_path_data} doesnt have a version.
                    Library and OS needs version numbers in its yaml.
                """)
                sys.exit(1)

            path_schema[yaml_data['type']][comp_name] = {version : dir_path}

    apps_dict = path_schema.get('apps', {})
    libs_dict = path_schema.get('library', {})

    for app_name in list(apps_dict.keys()):
        try:
            supported_proc_list, supported_os_list, description, dep_lib_list, _ = get_yaml_data(app_name, apps_dict[app_name]['vless'])
        except KeyError:
            supported_proc_list, supported_os_list, description, dep_lib_list, _ = get_yaml_data(app_name, apps_dict[app_name]['path'][0])
        if proc_ip_name in supported_proc_list:
            app_dict = {app_name : {'description': description, 'depends_libs': dep_lib_list}}
            if 'standalone' in supported_os_list:
                supported_app_dict[proc_name]['standalone'].update(app_dict)
            if "freertos10_xilinx" in supported_os_list:
                supported_app_dict[proc_name]['freertos'].update(app_dict)

    for lib_name in list(libs_dict.keys()):
        cur_lib_dict = libs_dict[lib_name]
        version_list = list(cur_lib_dict.keys())
        if 'path' in version_list:
            lib_dir = cur_lib_dict['path'][0]
            sorted_lib_dict = {version: cur_lib_dict[version] for version in version_list}
        else:
            version_list.sort(key = float, reverse = True)
            lib_dir = cur_lib_dict[version_list[0]]
            sorted_lib_dict = {version: cur_lib_dict[version] for version in version_list}

        supported_proc_list, supported_os_list, description, dep_lib_list, examples = get_yaml_data(lib_name, lib_dir,proc_ip_name,family,variant)
        if proc_ip_name in supported_proc_list:
            if 'path' in version_list:
                lib_dict = {lib_name : {'description': description, 'depends_libs': dep_lib_list, 'path': sorted_lib_dict['path'],'examples':examples}}
            else:
                lib_dict = {lib_name : {'description': description, 'depends_libs': dep_lib_list, 'versions': sorted_lib_dict,'examples':examples}}
            if 'standalone' in supported_os_list:
                supported_libs_dict[proc_name]['standalone'].update(lib_dict)
            if "freertos10_xilinx" in supported_os_list:
                supported_libs_dict[proc_name]['freertos'].update(lib_dict)


    with open(os.path.join(sdt.outdir, 'app_list.yaml'), 'w') as fd:
        fd.write(yaml.dump(supported_app_dict, sort_keys=False, indent=2, width=32768, Dumper=VerboseSafeDumper))

    with open(os.path.join(sdt.outdir, 'lib_list.yaml'), 'w') as fd:
        fd.write(yaml.dump(supported_libs_dict, sort_keys=False, indent=2, width=32768, Dumper=VerboseSafeDumper))

    return True


#/*
# * Copyright (c) 2023 - 2026 Advanced Micro Devices, Inc. All Rights Reserved.
# *
# * Author:
# *       Onkar Harsh <onkar.harsh@amd.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import os
import sys
import re
import glob
import yaml
from typing import Any, List, Optional, Dict, Union
import shutil
import logging
from lopper.log import _init, _warning

_init(__name__)

def to_cmakelist(pylist):
    cmake_list = ';'.join(pylist)
    cmake_list = '"{}"'.format(cmake_list)

    return cmake_list

def is_file(filepath: str, silent_discard: bool = True) -> bool:
    """Return True if the file exists Else returns False and raises Not Found Error Message.
    
    Args:
        filepath: File Path.
    Raises:
        FileNotFoundError: Raises exception if file not found.
    Returns:
        bool: True, if file is found Or False, if file is not found.
    """
   
    if os.path.isfile(filepath):
        return True
    elif not silent_discard:
        err_msg = f"No such file exists: {filepath}"
        raise FileNotFoundError(err_msg) from None
    else:
        return False

def is_dir(dirpath: str, silent_discard: bool = True) -> bool:
    """Checks if directory exists.
    
    Args:
        dirpath: Directory Path.
    Raises:
        ValueError (Exception): Raises exception if directory not found.
    Returns:
        bool: True, if directory is found Or False, if directory is not found.
    """

    if os.path.isdir(dirpath):
        return True
    elif not silent_discard:
        err_msg = f"No such directory exists: {dirpath}"
        raise ValueError(err_msg) from None
    else:
        return False

def get_base_name(fpath):
    """
    This api takes rel path or full path and returns base name
    Args:
        fpath: Path to get the base name from.
    Returns:
        string: Base name of the path
    """
    return os.path.basename(fpath.rstrip(os.path.sep))

def get_dir_path(fpath):
    """
    This api takes file path and returns it's directory path
    
    Args:
        fpath: Path to get the directory path from.
    Returns:
        string: Full Directory path of the passed path
    """
    return os.path.dirname(fpath.rstrip(os.path.sep))

def get_abs_path(fpath):
    """
    This api takes file path and returns it's absolute path
    Args:
        fpath: Path to get the absolute path from.
    Returns:
        string: Absolute location of the passed path
    """
    return os.path.abspath(fpath)

def load_yaml(filepath: str) -> Optional[dict]:
    """Read yaml file data and returns data in a dict format.
    
    Args:
        filepath: Path of the yaml file.
    Returns:
        dict: Return Python dict if the file reading is successful.
    """

    if is_file(filepath):
        try:
            with open(filepath) as f:
                data = yaml.safe_load(f)
            return data
        except Exception as e:
            _warning(f"{filepath} file reading failed: {e}")
            return {}
    else:
        return {}

def copy_file(src: str, dest: str, follow_symlinks: bool = False, silent_discard: bool = True) -> None:
    """
    copies the file from source to destination.
    Args:
        | src: source file path
        | dest: destination file path
        | follow_symlinks: maintain the symlink while copying
        | silent_discard: Dont raise exception if the source file doesnt exist 
    """
    is_file(src, silent_discard)
    shutil.copy2(src, dest, follow_symlinks=follow_symlinks)
    os.chmod(dest, 0o644)

def find_files(search_pattern, search_path):
    """
    This api find the files matching regex directories and returns absolute
    path of files, if file exists

    Args:
        | search_pattern: The regex pattern to be searched in file names
        | search_path: The directory that needs to be searched
    Returns:
        string: All the file paths that matches the pattern in the searched path.

    """

    return glob.glob(f"{search_path}{os.path.sep}{search_pattern}")

def parse_depends_schema(depends_dict):
    """
    Parse the depends dictionary to determine validation mode and extract driver requirements.

    Supports three formats:
    1. Traditional flat format: {'driver1': [props], 'driver2': [props]} - all required
    2. oneOf format: {'oneOf': [{'driver1': [props]}, {'driver2': [props]}]} - at least one required
    3. anyOf format: {'anyOf': [{'driver1': [props]}, {'driver2': [props]}]} - at least one required

    Args:
        depends_dict: The depends dictionary from YAML schema

    Returns:
        tuple: (validation_mode, driver_requirements_dict)
               validation_mode: 'all', 'oneOf', or 'anyOf'
               driver_requirements_dict: dict mapping driver names to their required properties
    """
    if not depends_dict:
        return ('all', {})

    # Check for oneOf and anyOf formats
    for mode in ('oneOf', 'anyOf'):
        if mode in depends_dict:
            driver_reqs = {}
            for item in depends_dict[mode]:
                if isinstance(item, dict):
                    driver_reqs.update(item)
            return (mode, driver_reqs)
    # Traditional flat format - all drivers required
    driver_reqs = {k: v for k, v in depends_dict.items() if k not in ['condition']}
    return ('all', driver_reqs)
    
def log_setup(options):
  
    """
    Sets up the log level based on the verbosity given by the user.
    
    Args:
        options (dict): Dictionary containing command-line arguments.
    
    Returns:
        int: Logging level.
    """
    verbose = [i for i in options.get("args", []) if i.startswith('-v')]
    verbose_level = 1 if verbose else 0 
    # Adjust logging level based on verbose level
    level = logging.DEBUG if verbose_level >= 1 else logging.WARNING
    #print(f"[LOG_SETUP] Verbose level = {verbose_level}, Final level = {level}")
    return level

def run_exec(yaml_condition, proc_ip_name, family, variant=None, return_list="examples", yaml_file=""):
    """
    Executes a Python condition string in a restricted local scope and returns a specified list.

    Args:
        yaml_condition (str): Python code to execute, typically from a YAML file.
        proc_ip_name (str): Name of the processor IP available in the local scope as 'proc'.
        family (str): Platform family available in the local scope as 'platform'.
        variant (str, optional): Variant available in the local scope as 'variant'.
        return_list (str, optional): Name of the list to return from the local scope. Defaults to "examples".
        yaml_file (str, optional): YAML file name for error reporting.

    Returns:
        list: The list named by `return_list` from the local scope after executing the condition.

    Notes:
        Any exceptions during execution are caught and logged as warnings.
    """
    local_scope = {
        "proc": proc_ip_name,
        "platform": family,
        "variant": variant,
        return_list: []
    }
    try:
        exec(yaml_condition, {"__builtins__": {}}, local_scope)
    except Exception as e:
        _warning(f"The condition in the {yaml_file} file has failed. -> {e}")
    finally:
        return local_scope[return_list]
